/* GGL - GPUL Game Library
 * Copyright (C) 1998 G.P.U.L. (Group of programmers and users of Linux)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <ggl/ggltimeout.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>

static struct {
    GList *ggl_timeouts;
    int ggl_usec;           /* Greatest common divisor of all usec's */
    int ggl_timeout_locked; /* The timestamp is >=1 usec, then we don't need semaphores */
} ggl_alarm={NULL,0,0};

int gcd(int a, int b)
{
  int res=1;
  int i;
  if (a==0) return b;
  if (b==0) return a;
  for (i=2;i<=a && i<=b;i++)
  {
    if ((a%i ==0) && (b%i ==0)) res=i;
  }
  
  return res;
}

void ggltimeout_foreach(GglTimeout *timeout)
{
    GList *node;
    GglCallback *callback;
    if (!timeout->active) return;
    if (timeout->curtick==0)
    {
      for (node=timeout->callbacks;node!=NULL;node=node->next)
      {
        callback=(GglCallback*)node->data;
        callback->callback(callback->data);
      }
    }
    timeout->curtick=(timeout->curtick+1)%timeout->ticks;
}

void ggltimeout_handler(int i)
{
  signal(SIGALRM,ggltimeout_handler);
  if (!ggl_alarm.ggl_timeout_locked)
  {
    ggl_alarm.ggl_timeout_locked=1;
    g_list_foreach(ggl_alarm.ggl_timeouts,(GFunc)ggltimeout_foreach,NULL);
    ggl_alarm.ggl_timeout_locked=0;
  }    
}

void ggltimeout_start()
{
  struct itimerval it;
  if (ggl_alarm.ggl_usec==0) {
    g_warning("GGL_TIMEOUT: Can't start timeout, usec=0\n");
    return;
  }
  it.it_interval.tv_sec=ggl_alarm.ggl_usec/1000000;
  it.it_interval.tv_usec=ggl_alarm.ggl_usec%1000000;
  it.it_value.tv_sec=0;
  it.it_value.tv_usec=1;
  setitimer(ITIMER_REAL,&it,0);
  signal(SIGALRM,ggltimeout_handler);
}

void ggltimeout_install(GglTimeout *timeout)
{
  int new_usec;
  GList *timeouts=ggl_alarm.ggl_timeouts;
  GglTimeout *my_timeout;
  new_usec=gcd(ggl_alarm.ggl_usec,timeout->usec);
  for (timeouts=ggl_alarm.ggl_timeouts;timeouts!=NULL;timeouts=timeouts->next)
    {
      my_timeout=(GglTimeout*)timeouts->data;
      my_timeout->ticks*=ggl_alarm.ggl_usec/new_usec;
      my_timeout->curtick*=ggl_alarm.ggl_usec/new_usec;
    }
  ggl_alarm.ggl_usec=new_usec;
  ggl_alarm.ggl_timeouts=g_list_append(ggl_alarm.ggl_timeouts,timeout);
  timeout->ticks=timeout->usec/ggl_alarm.ggl_usec;
  timeout->curtick=0;
  ggltimeout_start(); /* Reset the timer */
}

void ggltimeout_remove(GglTimeout *timeout)
{
  /* FIXME: Recalculate the gcd and the timeout's ticks */
  ggl_alarm.ggl_timeouts=g_list_remove(ggl_alarm.ggl_timeouts,timeout);
}

void ggl_timeout_init(GglTimeout *timeout, int sec, int usec)
{
  timeout-> id=GGLTIMEOUT;
  timeout-> usec=usec+sec*1000000;
  timeout-> callbacks=NULL;
  timeout-> ticks=1;
  timeout-> curtick=0;
  timeout-> active=0;
}

GglTimeout* ggl_timeout_new(int sec, int usec)
{
  GglTimeout *timeout=(GglTimeout *)malloc(sizeof(GglTimeout));
  ggl_base_set_size(GGL_BASE(timeout),sizeof(GglTimeout));
  
  ggl_timeout_init(timeout,sec,usec);
  ggltimeout_install(timeout);
  
  return timeout;
}

void ggl_timeout_dispose(GglTimeout *timeout)
{
  ggl_timeout_finalize(timeout);
  free(timeout);
}
					 
void ggl_timeout_finalize(GglTimeout *timeout)
{
  ggltimeout_remove(timeout);
}

void ggl_timeout_add_callback(GglTimeout *timeout, GglCallback callback)
{
  GglCallback *my_callback=(GglCallback *)malloc(sizeof(GglCallback));
  memcpy(my_callback,&callback,sizeof(GglCallback));
  timeout->callbacks=g_list_append(timeout->callbacks, my_callback);
}

void ggl_timeout_start(GglTimeout *timeout)
{
  timeout->active=1;
}
