/* GGL - GPUL Game Library
 * Copyright (C) 1998 G.P.U.L. (Group of programmers and users of Linux)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <ggl/gglgc.h>
#include <ggl/gglsprite.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifndef PI
  #define PI 3.14159265356
#endif

typedef enum {
    UM_POSITION,
    UM_VISIBILITY
} UserModif;

int  gglsprite_update(GglSprite*);
void gglsprite_draw (GglSprite*, GglVisual*, int, int, GList*);
void gglsprite_area(GglSprite*, GglRectangle*);

/************************************************************/
void ggl_sprite_init(GglSprite *sprite)
{
    ggl_graphic_init(GGL_GRAPHIC(sprite));

    sprite-> id=GGLSPRITE;
    sprite-> anims=NULL;
    sprite-> speed_x=0;
    sprite-> speed_y=0;
    sprite-> cur_anim=NULL;

    GGL_GRAPHIC(sprite)->draw=(void*)gglsprite_draw;
    GGL_GRAPHIC(sprite)->change=(void*)gglsprite_update;
    GGL_GRAPHIC(sprite)->get_area=(void*)gglsprite_area;
}

/************************************************************/
GglSprite* ggl_sprite_new (void)
{
    GglSprite *sprite=(GglSprite *)malloc(sizeof(GglSprite));
    ggl_base_set_size(GGL_BASE(sprite),sizeof(GglSprite));

    ggl_sprite_init(sprite);

    return sprite;
}

/************************************************************/
void ggl_sprite_finalize(GglSprite *sprite)
{
    ggl_graphic_finalize(GGL_GRAPHIC(sprite));
    g_list_free(sprite->anims);
}

/************************************************************/
void ggl_sprite_dispose(GglSprite *sprite)
{
    ggl_sprite_finalize(sprite);
    free(sprite);
}

/************************************************************/
void ggl_sprite_append (GglSprite *sprite, GglAnim *anim)
{
    sprite->anims=g_list_append(sprite->anims, anim);
    if (sprite->cur_anim==NULL)
	sprite->cur_anim=ggl_descriptor_new(sprite->anims);
}

/************************************************************/
void ggl_sprite_prepend(GglSprite *sprite, GglAnim *anim)
{
    g_list_prepend(sprite->anims, anim);
    ggl_descriptor_dispose(sprite->cur_anim);
    if (sprite->cur_anim==NULL)
	sprite->cur_anim=ggl_descriptor_new(sprite->anims);
}

/************************************************************/
void ggl_sprite_remove(GglSprite *sprite, GglAnim *anim)
{
    sprite->anims=g_list_remove(sprite->anims, anim);
    ggl_descriptor_dispose(sprite->cur_anim);
    sprite->cur_anim=ggl_descriptor_new(sprite->anims);
}

/************************************************************/
void ggl_sprite_insert (GglSprite *sprite, GglAnim *anim, int position)
{
    sprite->anims=g_list_insert(sprite->anims,anim,position);
    if (sprite->cur_anim==NULL)
	sprite->cur_anim=ggl_descriptor_new(sprite->anims);
}

/************************************************************/
int ggl_sprite_set_anim(GglSprite *sprite, GglAnim *anim)
{
    GList *item = g_list_find(sprite->anims, anim);
    if (item)
    {
        ggl_descriptor_dispose(sprite->cur_anim);
        sprite->cur_anim=ggl_descriptor_new(item);
        GGL_GRAPHIC(sprite)->modified=1;
        return 0;
    }else{
        return -1;
    }
        
}

/************************************************************/
void ggl_sprite_set_anim_nth(GglSprite *sprite, int anim)
{
    ggl_descriptor_dispose(sprite->cur_anim);
    sprite->cur_anim=ggl_descriptor_new(g_list_nth(sprite->anims,anim));
    GGL_GRAPHIC(sprite)->modified=1;
}

/************************************************************/
GglDescriptor *ggl_sprite_get_anim(GglSprite *sprite)
{
    return sprite->cur_anim;
}

/************************************************************/
void ggl_sprite_set_frame(GglSprite *sprite, GglDescriptor *descriptor)
{
    GglAnim *anim;

    anim=(GglAnim*)ggl_descriptor_data(sprite->cur_anim);
    ggl_anim_set_frame(anim,descriptor);
    GGL_GRAPHIC(sprite)->modified=1;
}

/************************************************************/
void ggl_sprite_set_frame_nth(GglSprite *sprite, int frame)
{
    GglAnim *anim;

    anim=(GglAnim*)ggl_descriptor_data(sprite->cur_anim);
    ggl_anim_set_frame_nth(anim,frame);
    GGL_GRAPHIC(sprite)->modified=1;
}

/************************************************************/
GglDescriptor* ggl_sprite_get_frame(GglSprite *sprite)
{
    GglAnim *anim;

    anim=(GglAnim*)ggl_descriptor_data(sprite->cur_anim);
    return ggl_anim_get_frame(anim);
}

/************************************************************/
void gglsprite_draw (GglSprite *sprite, GglVisual *visual, int ofs_x, int ofs_y, GList *overlaps)
{
    GglAnim *anim;
    
    if (sprite->anims!=NULL)
    {
	GList *item;
	int x,y,clipx0,clipy0,clipx1,clipy1;

	anim=(GglAnim*)ggl_descriptor_data(sprite->cur_anim);
	x=(int)GGL_GRAPHIC(sprite)->x+ofs_x;
	y=(int)GGL_GRAPHIC(sprite)->y+ofs_y;
	ggl_gc_get_clipping(visual,&clipx0,&clipy0,&clipx1,&clipy1);

	for (item=overlaps;item!=NULL;item=item->next)
	{
	    GglRectangle *r=(GglRectangle*)item->data;
	    int x0,y0,x1,y1;
	    x0=r->x;
	    y0=r->y;
	    x1=r->x+r->width;
	    y1=r->y+r->height;
	    if (x0<0) {x0=0;}
	    if (y0<0) {y0=0;}
	    if (x1>=visual->width) x1=visual->width-1;
	    if (y1>=visual->height) y1=visual->height-1;
	    if (x0<visual->width && x1>=0 && y0<visual->height && y0>=0 && x0!=x1 && y0!=y1)
	    {
		ggl_gc_set_clipping(visual,x0,y0,x1,y1);
		ggl_anim_draw(anim,visual,x,y);
	    }
	}
	ggl_gc_set_clipping(visual,clipx0,clipy0,clipx1,clipy1);
    }
}

/************************************************************/
int gglsprite_update(GglSprite *sprite)
{
    GglAnim *anim;

    if (sprite->anims!=NULL)
    {
	anim=(GglAnim*)ggl_descriptor_data(sprite->cur_anim);

	if ( ggl_anim_length(anim) > 1 )
	{
	    ggl_anim_set_next_frame(anim);
	    GGL_GRAPHIC(sprite)->modified=1;
	}

	if (GGL_GRAPHIC(sprite)->user_modif & MODIF_POSITION)
	{
	    if ((int)GGL_GRAPHIC(sprite)->x!=(int)GGL_GRAPHIC(sprite)->m_x
		|| (int)GGL_GRAPHIC(sprite)->y!=(int)GGL_GRAPHIC(sprite)->m_y)
	    {
		GGL_GRAPHIC(sprite)->modified = 1;
	    }
	    GGL_GRAPHIC(sprite)->x=GGL_GRAPHIC(sprite)->m_x;
	    GGL_GRAPHIC(sprite)->y=GGL_GRAPHIC(sprite)->m_y;
	    GGL_GRAPHIC(sprite)->user_modif &= ~MODIF_POSITION;
	}
	
	if (GGL_GRAPHIC(sprite)->user_modif & MODIF_VISIBLE)
	{
	    GGL_GRAPHIC(sprite)->modified = 1;
	    GGL_GRAPHIC(sprite)->visible=GGL_GRAPHIC(sprite)->m_visible;
	    GGL_GRAPHIC(sprite)->user_modif &= ~MODIF_VISIBLE;
	}
	
	if (sprite->speed_x || sprite->speed_y)
	{
	    double oldx,oldy,newx,newy;
	    oldx=GGL_GRAPHIC(sprite)->x;
	    oldy=GGL_GRAPHIC(sprite)->y;
	    newx=oldx+sprite->speed_x;
	    newy=oldy+sprite->speed_y;
	    if ((int)oldx!=(int)newx || (int)oldy!=(int)newy)
	    {
		GGL_GRAPHIC(sprite)->modified=1;
	    }
	    GGL_GRAPHIC(sprite)->x=newx;
	    GGL_GRAPHIC(sprite)->y=newy;
	}
    }
    return GGL_GRAPHIC(sprite)->modified;
}

/************************************************************/
void gglsprite_area(GglSprite *sprite, GglRectangle *area)
{
    GglAnim *anim;
    short x_hook,y_hook;
    if (sprite->anims!=NULL)
    {
	anim=(GglAnim*)ggl_descriptor_data(sprite->cur_anim);

        if (anim->frames) /* Must change this condition, its ugly */
        {
	    ggl_anim_cur_frame_hook(anim,&x_hook,&y_hook);
	    area->x=(int)GGL_GRAPHIC(sprite)->x-x_hook;
	    area->y=(int)GGL_GRAPHIC(sprite)->y-y_hook;
	    ggl_anim_cur_frame_size(anim,&(area->width),&(area->height));
	}else{
	    area->x=0;area->y=0;area->width=0;area->height=0;
	}
    }else{
	area->x=0;area->y=0;area->width=0;area->height=0;
    }
}


/************************************************************/
void ggl_sprite_move (GglSprite *sprite, float x, float y)
{
    GGL_GRAPHIC(sprite) -> m_x = x;
    GGL_GRAPHIC(sprite) -> m_y = y;
    GGL_GRAPHIC(sprite) -> user_modif |= MODIF_POSITION;
}

/************************************************************/
void ggl_sprite_position (GglSprite *sprite, float *x, float *y)
{
    if (GGL_GRAPHIC(sprite)->user_modif)
    {
	*x=GGL_GRAPHIC(sprite)->m_x;
	*y=GGL_GRAPHIC(sprite)->m_y;
    }else{
	*x=GGL_GRAPHIC(sprite)->x;
	*y=GGL_GRAPHIC(sprite)->y;
    }
}

/************************************************************/
void ggl_sprite_set_speed (GglSprite *sprite, float sx, float sy)
{
    sprite->speed_x=sx;
    sprite->speed_y=sy;
}

/************************************************************/
void ggl_sprite_get_speed (GglSprite *sprite, float *sx, float *sy)
{
    *sx=sprite->speed_x;
    *sy=sprite->speed_y;
}


/************************************************************/
void gglsprite_foreach_clone(GglAnim *anim, GglSprite *new_sprite)
{
    GglAnim *new_anim;
    if (anim!=NULL)
    {
	new_anim=ggl_anim_clone(anim);
	ggl_sprite_append(new_sprite,new_anim);
    }
}

/************************************************************/
GglSprite *ggl_sprite_clone (GglSprite *sprite)
{
    GglSprite *new=malloc(sizeof(GglSprite));
    /* The new sprite will share the same frames, eating less memory :) */
    memcpy(new,sprite,sizeof(GglSprite));
    /* Now we create a new animation list */
    new->anims=NULL;
    /* And copy the animations of the other sprite */
    g_list_foreach(sprite->anims,(GFunc)gglsprite_foreach_clone,new);
    new->cur_anim=ggl_descriptor_new(new->anims);
    return new;
}

/************************************************************/
void ggl_sprite_cur_frame_size (GglSprite *sprite, int *width, int *height)
{
    GglAnim *anim;

    anim=(GglAnim*)ggl_descriptor_data(sprite->cur_anim);
    ggl_anim_cur_frame_size(anim,width,height);
}


/************************************************************/
int gglsprite_contour_collision(GglSprite *first, GglSprite *second)
{
    int x1l,y1u,x1r,y1d; // bounds of the first object
    int x2l,y2u,x2r,y2d; // bounds of the second object

    x1l=GGL_GRAPHIC(first)->x;
    y1u=GGL_GRAPHIC(first)->y;
    ggl_sprite_cur_frame_size(first,&x1r,&y1d);
    x1r+=x1l;
    y1d+=y1u;

    x2l=GGL_GRAPHIC(second)->x;
    y2u=GGL_GRAPHIC(second)->y;
    ggl_sprite_cur_frame_size(second,&x2r,&y2d);
    x2r+=x2l;
    y2d+=y2u;

    if (x2l>=x1r || y2u>=y1d || x2r<=x1l || y2d<=y1u) return 0;
    else {
	return 1;
    }
}

/************************************************************/
int ggl_sprite_collision(GglSprite *first, GglSprite *second, float *angle)
{
    /* NOTE: angle can be NULL, indicating that the caller routine doesn't want to
     * know the angle */
    if (!gglsprite_contour_collision(first,second)) return 0;
    else
    {
	GglFrame *frame1,*frame2;
	frame1=(GglFrame*)ggl_descriptor_data(ggl_anim_get_frame((GglAnim*)ggl_descriptor_data(first->cur_anim)));
	frame2=(GglFrame*)ggl_descriptor_data(ggl_anim_get_frame((GglAnim*)ggl_descriptor_data(second->cur_anim)));

	return ggl_frame_collision(frame1,frame2,GGL_GRAPHIC(first)->x,
			GGL_GRAPHIC(first)->y,GGL_GRAPHIC(second)->x,
			GGL_GRAPHIC(second)->y,angle);
    }
}

/************************************************************************/
/** MATRIX ALGEBRA STUFF **/
/****/

inline static void swapf(float m[2][4],int f1, int f2)
{
    int i;
    for (i=0;i<4;i++)
    {
	float tmp=m[f1][i];
	m[f1][i]=m[f2][i];
	m[f2][i]=tmp;
    }
}

static void tri_sup(float m[2][4])
{
    int i,j,k;
    float abajo,arriba;
    for(i=0;i<2-1;i++)
    {
	for(j=i+1;j<2;j++)
	{
	    abajo=m[j][i];
	    arriba=m[i][i];
	    if (m[i][i]==0) swapf(m,i,j);
	    else for(k=0;k<4;k++)
		m[j][k]-=(m[i][k]*abajo)/arriba;
	}
    }
}

static void tri_inf(float m[2][4])
{
int i,j,k;
float abajo,arriba;
for(i=2-1;i>0;i--)
  {
  for(j=i-1;j>=0;j--)
     {
     arriba=m[j][i];
     abajo=m[i][i]; 
     if (m[i][i]==0) swapf(m,i,j);
     else for(k=0;k<4;k++)
                m[j][k]-=(m[i][k]*arriba)/abajo;
     }
  }   
}


/************************************************************/
void ggl_sprite_bound(GglSprite *sprite, float angle)
{
    int i;
    float base[2][2];
    float base_inv[2][4];
    float newspeed_x,newspeed_y;
    if (angle>PI)
    {
	if (angle<2*PI)
	    angle-=PI;
	else /* Angle is Bigger than 2*PI :-m */
	{
	    int times=angle/PI;
	    angle-=times*PI;
	}
    }

    /* The easy cases... */
    if (((int)(angle*100))==0)
    {
	sprite->speed_y=-sprite->speed_y;
	return;
    }
    if (((int)(angle*100))==((int)(PI/2*100)))
    {
	sprite->speed_x=-sprite->speed_x;
	return;
    }

    /* And now the general case */
    base[0][0]=cos(angle);
    base[0][1]=sin(angle);
    base[1][0]=cos(angle+PI/2);
    base[1][1]=sin(angle+PI/2);

    base_inv[0][0]=base[0][0];
    base_inv[0][1]=base[0][1];
    base_inv[1][0]=base[1][0];
    base_inv[1][1]=base[1][1];
    base_inv[0][2]=1;
    base_inv[0][3]=0;
    base_inv[1][2]=0;
    base_inv[1][3]=1;

    /* Inverse matrix */
    tri_sup(base_inv); // Zeros down
    tri_inf(base_inv); // Zeros up

    for (i=0;i<2;i++)
    {
	float val=base_inv[i][i];
	if (val!=0)
	{
	    int j;
	    if (val!=1)
		for (j=0;j<4;j++) base_inv[i][j]/=val;
	}
	else g_warning("ggl_sprite_bound: Base Change Matrix hasn't inverse!\n");
    }
#ifdef DEBUG
    for (i=0;i<2;i++)
    {
	int j;
	for (j=0;j<2;j++)
	{
	    printf("%f ",base[i][j]);
	}printf("|");
	for (j=0;j<4;j++)
	{
	    printf("%f ",base_inv[i][j]);
	}printf("\n");
    }
    printf("ORIGINAL:(%f,%f)\n",sprite->speed_x,sprite->speed_y);
#endif
    /* Now the right part of base_inv has the inverse of the base matrix. */
    /* First of all will change speed to the new base */
    /* must use real y coordinate instead of screen y coordinate */
    sprite->speed_y=-sprite->speed_y;
    /* Now can calculate the rest*/
    newspeed_x=sprite->speed_x*base_inv[0][2] + sprite->speed_y*base_inv[1][2];
    newspeed_y=sprite->speed_x*base_inv[0][3] + sprite->speed_y*base_inv[1][3];

#ifdef DEBUG
    printf("NEWS:(%f,%f)\n",newspeed_x,newspeed_y);
#endif
    /* Now invert y coordinate on new axis*/
    newspeed_y=-newspeed_y;

    /* And now return to the canonical base */
    sprite->speed_x=newspeed_x*base[0][0] +newspeed_y*base[1][0];
    sprite->speed_y=newspeed_x*base[0][1] +newspeed_y*base[1][1];
    /* Must return to screen y coordinate (inverse of the real one) */
    sprite->speed_y=-sprite->speed_y;
#ifdef DEBUG
    printf("LAST:(%f,%f)\n",sprite->speed_x,sprite->speed_y);
#endif
}
